(*
 * Graph: generic graph library
 * Copyright (C) 2004
 * Sylvain Conchon, Jean-Christophe Filliatre and Julien Signoles
 * 
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License version 2, as published by the Free Software Foundation.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * See the GNU Library General Public License version 2 for more details
 * (enclosed in the file LGPL).
 *)

(* $Id: flow.ml,v 1.12 2004/02/05 10:04:02 signoles Exp $ *)

module type FLOW = sig
  type label
  type t
  val max_capacity : label -> t
  val min_capacity : label -> t
  val flow : label -> t
  val add : t -> t -> t
  val remove : t -> t -> t
  val zero : t
  val max : t
  val min : t
  val compare : t -> t -> int
end

module type G = sig
  type t
  module V : Sig.HASHABLE
  module E : sig
    type t
    type label
    val src : t -> V.t
    val dst : t -> V.t
    val label : t -> label
  end
  val iter_succ_e : (E.t -> unit) -> t -> V.t -> unit
  val iter_pred_e : (E.t -> unit) -> t -> V.t -> unit
end

module MaxFlow(G: G)(F: FLOW with type label = G.E.label) = 
struct

  module Mark = struct
    module H = Hashtbl.Make(G.V)
    type mark = Plus | Minus

    let marked = H.create 997
    let unvisited = Queue.create ()

    let clear () = H.clear marked

    let mem = H.mem marked

    let set s e tag = 
      assert (not (mem s));
      H.add marked s (e, tag);
      Queue.add s unvisited

    let get s : G.E.t * mark =
      let e, tag = H.find marked s in
      (match e with None -> assert false | Some e -> e), tag

    exception Empty = Queue.Empty
    let next () = Queue.pop unvisited
  end

  module Result = struct
    module H = Hashtbl.Make(Util.HTProduct(G.V)(G.V))
    let result = H.create 997

    let find = H.find result

    let flow e = 
      let key = (G.E.src e, G.E.dst e) in
      try
	find key
      with Not_found ->
	let f = F.flow (G.E.label e) in
	H.add result key f;
	f

    let change op e f =
      let key = (G.E.src e, G.E.dst e) in
      try
	H.replace result key (op (find key) f);
      with Not_found ->
	assert false

    let grow = change F.add
    let reduce = change F.remove
  end

  module H = Result.H
		     
  let is_full e = 
    F.compare (F.max_capacity (G.E.label e)) (Result.flow e) = 0

  let is_empty e =
    F.compare F.min (Result.flow e) = 0

  let set_flow s t a =
    let rec loop t =
      if not (G.V.equal s t) then
	let e, tag = Mark.get t in
	(match tag with 
	   | Mark.Plus -> Result.grow 
	   | Mark.Minus -> Result.reduce) e a;
	loop (G.E.src e)
    in
    loop t

  let grow_flow s t a =
    let rec loop u b =
      if G.V.equal s u then begin
	set_flow s t b;
	F.add a b
      end else
	let e, tag = Mark.get u in
	let l = G.E.label e in
	let b = match tag with
	  | Mark.Plus -> min b (F.remove (F.max_capacity l) (Result.flow e))
	  | Mark.Minus -> min b (F.remove (Result.flow e) (F.min_capacity l))
	in
	loop (G.E.src e) b
    in
    loop t F.max

  let ford_fulkerson g s t =
    let succ s = 
      G.iter_succ_e
	(fun e ->
	   assert (s = G.E.src e);
	   let t = G.E.dst e in
	   if not (Mark.mem t || is_full e) then Mark.set t (Some e) Mark.Plus)
	g s
    in
    let prev s = 
      G.iter_pred_e
	(fun e ->
	   assert (s = G.E.dst e);
	   let t = G.E.src e in
	   if not (Mark.mem t || is_empty e) then
	     Mark.set t (Some e) Mark.Minus)
	g s
    in
    let internal_loop a =
      try
	while true do let s = Mark.next () in succ s; prev s done;
	assert false
      with Mark.Empty ->
	if Mark.mem t then grow_flow s t a else a
    in
    let rec external_loop a =
      Mark.clear ();
      Mark.set s None Mark.Plus;
      let a' = internal_loop a in
      if a = a' then a else external_loop a'
    in
    let a = external_loop F.zero in
    Result.result, a

end
