(*
 * Graph: generic graph library
 * Copyright (C) 2004
 * Sylvain Conchon, Jean-Christophe Filliatre and Julien Signoles
 * 
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License version 2, as published by the Free Software Foundation.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * See the GNU Library General Public License version 2 for more details
 * (enclosed in the file LGPL).
 *)

(* $Id: sig_pack.mli,v 1.6 2004/02/05 10:08:57 filliatr Exp $ *)

(** Immediate access to the library.
    Signature [S] gathers an imperative implementation and all algorithms into 
    a single module. 
    Vertices and edges are labeled with integers. *)

module type S = sig

  (** abstract type of graphs *)

  type t

  (** vertices are labeled with integers *)

  module V : sig
    type t 
    val compare : t -> t -> int 
    val hash : t -> int 
    val equal : t -> t -> bool
    type label = int
    val create : label -> t
    val label : t -> label
  end

  (** edges are labeled with integers *)

  module E : sig
    type t
    val compare : t -> t -> int
    type label = int
    val create : V.t -> label -> V.t -> t
    val src : t -> V.t
    val dst : t -> V.t
    val label : t -> label
  end

  (** graph constructors and destructors *)

  val create : unit -> t
  val copy : t -> t

  val add_vertex : t -> V.t -> unit
  val remove_vertex : t -> V.t -> unit

  val add_edge : t -> V.t -> V.t -> unit
  val add_edge_e : t -> E.t -> unit
  val remove_edge : t -> V.t -> V.t -> unit
  val remove_edge_e : t -> E.t -> unit

  (** Vertices contains integers marks, which can be set or used by some 
      algorithms (see for instance module [Marking] below) *)

  module Mark : sig
    val clear : t -> unit
    val get : V.t -> int
    val set : V.t -> int -> unit
  end

  (** is this an implementation of directed graphs? *)

  val is_directed : bool

  (** size functions *)

  val is_empty : t -> bool
  val nb_vertex : t -> int
  val nb_edges : t -> int

  (** membership functions *)

  val mem_vertex : t -> V.t -> bool
  val mem_edge : t -> V.t -> V.t -> bool
  val mem_edge_e : t -> E.t -> bool

  (** successors and predecessors of a vertex *)

  val succ : t -> V.t -> V.t list
  val pred : t -> V.t -> V.t list
  val in_degree : t -> V.t -> int
  val out_degree : t -> V.t -> int

  (** iter/fold on all vertices/edges of a graph *)

  val iter_vertex : (V.t -> unit) -> t -> unit
  val iter_edges : (V.t -> V.t -> unit) -> t -> unit
  val fold_vertex : (V.t -> 'a -> 'a) -> t  -> 'a -> 'a
  val fold_edges : (V.t -> V.t -> 'a -> 'a) -> t -> 'a -> 'a

  (** iter/fold on all successor/predecessor of a vertex. *)

  val iter_succ : (V.t -> unit) -> t -> V.t -> unit
  val iter_pred : (V.t -> unit) -> t -> V.t -> unit
  val fold_succ : (V.t -> 'a -> 'a) -> t -> V.t -> 'a -> 'a
  val fold_pred : (V.t -> 'a -> 'a) -> t -> V.t -> 'a -> 'a

  (** map iterator on vertex *)

  val map_vertex : (V.t -> V.t) -> t -> t

  (** labeled edges going from/to a vertex *)

  val succ_e : t -> V.t -> E.t list
  val pred_e : t -> V.t -> E.t list

  (** iter/fold on all labeled edges of a graph *)

  val iter_edges_e : (E.t -> unit) -> t -> unit
  val fold_edges_e : (E.t -> 'a -> 'a) -> t -> 'a -> 'a

  (** iter/fold on all edges going from/to a vertex *)

  val iter_succ_e : (E.t -> unit) -> t -> V.t -> unit
  val fold_succ_e : (E.t -> 'a -> 'a) -> t -> V.t -> 'a -> 'a
  val iter_pred_e : (E.t -> unit) -> t -> V.t -> unit
  val fold_pred_e : (E.t -> 'a -> 'a) -> t -> V.t -> 'a -> 'a

  (** [vertex g i] returns a vertex of label [i] in [g]. The behaviour is
      unspecified if [g] has several vertices with label [i]. 
      Note: this function is inefficient (linear in the number of vertices);
      you should better keep the vertices as long as you create them. *)
  val find_vertex : t -> int -> V.t

  (** basic operations *)

  val transitive_closure : ?reflexive:bool -> t -> t
      (** [transitive_closure ?reflexive g] returns the transitive closure 
	of [g] (as a new graph). Loops (i.e. edges from a vertex to itself) 
	are added only if [reflexive] is [true] (default is [false]). *)

  val add_transitive_closure : ?reflexive:bool -> t -> t
      (** [add_transitive_closure ?reflexive g] replaces [g] by its
	transitive closure. Meaningless for persistent implementations
	(then acts as [transitive_closure]). *)

  val mirror : t -> t
      (** [mirror g] returns a new graph which is the mirror image of [g]:
	each edge from [u] to [v] has been replaced by an edge from [v] to [u].
	For undirected graphs, it simply returns of copy of [g]. *)

  val complement : t -> t
      (** [complement g] builds a new graph which is the complement of [g]:
	each edge present in [g] is not present in the resulting graph and
	vice-versa. Edges of the returned graph are unlabeled. *)

  (** Traversal *)

  (** Depth-first search *)
  module Dfs : sig
    val iter : ?pre:(V.t -> unit) -> 
               ?post:(V.t -> unit) -> t -> unit
      (** [iter pre post g] visits all nodes of [g] in depth-first search, 
	  applying [pre] to each visited node before its successors,
	  and [post] after them. Each node is visited exactly once. *)
    val prefix : (V.t -> unit) -> t -> unit
      (** applies only a prefix function *)
    val postfix : (V.t -> unit) -> t -> unit
      (** applies only a postfix function *)

    (** Same thing, but for a single connected component *)

    val iter_component : 
               ?pre:(V.t -> unit) -> 
               ?post:(V.t -> unit) -> t -> V.t -> unit
    val prefix_component : (V.t -> unit) -> t -> V.t -> unit
    val postfix_component : (V.t -> unit) -> t -> V.t -> unit

    val has_cycle : t -> bool
  end

  (** Breadth-first search *)
  module Bfs : sig
    val iter : (V.t -> unit) -> t -> unit
    val iter_component : (V.t -> unit) -> t -> V.t -> unit
  end

  (** Graph traversal with marking *)
  module Marking : sig
    val dfs : t -> unit
    val has_cycle : t -> bool
  end

  (** classic graphs *)
  module Classic : sig
    val divisors : int -> t
	(** [divisors n] builds the graph of divisors. 
	  Vertices are integers from [2] to [n]. [i] is connected to [j] if
          and only if [i] divides [j]. Raises [Invalid_argument] is [n < 2]. *)

    val de_bruijn : int -> t
	(** [de_bruijn n] builds the de Bruijn graph of order [n].
	  Vertices are bit sequences of length [n] (encoded as their
	  interpretation as binary integers). The sequence [xw] is connected
	  to the sequence [wy] for any bits [x] and [y] and any bit sequence 
          [w] of length [n-1]. 
	  Raises [Invalid_argument] is [n < 1] or [n > Sys.word_size-1]. *)

    val vertex_only : int -> t
	(** [vertex_only n] builds a graph with [n] vertices and no edge. *)

    val full : ?self:bool -> int -> t
	(** [full n] builds a graph with [n] vertices and all possible edges.
	  The optional argument [self] indicates if loop edges should be added
          (default value is [true]). *)
  end

  (** random graphs *)
  module Rand : sig
    val random : v:int -> e:int -> t
	(** [random v e] generates a random with [v] vertices and [e] edges. *)
  end

  (** strongly connected components *)
  val scc : t -> (V.t -> int)

  (** Dijkstra's shortest path algorithm. Weights are the labels. *)
  val shortest_path : t -> V.t -> V.t -> E.t list * int

  (** Ford Fulkerson *)
  module H : Hashtbl.S with type key = V.t * V.t
  val ford_fulkerson : t -> V.t -> V.t -> int H.t * int

  (** topological order *)
  module Topological : sig
    val fold : (V.t -> 'a -> 'a) -> t -> 'a -> 'a
    val iter : (V.t -> unit) -> t -> unit
  end

  (** Kruskal algorithm *)
  val spanningtree : t -> E.t list

  (** DOT output *)
  val dot_output : t -> string -> unit 

end
